import numpy as np
import keras.utils


class MockDataGenerator(keras.utils.Sequence):
    def __init__(
        self,
        num_images: int,
        input_frames: int = 1,
        output_frames: int = 1,
        batch_size: int = 8,
        input_height: int = 256,
        input_width: int = 256,
        output_height: int = 256,
        output_width: int = 256,
    ):
        self.input_height = input_height
        self.input_width = input_width
        self.output_height = output_height
        self.output_width = output_width
        self.input_frames = input_frames
        self.output_frames = output_frames
        self.batch_size = batch_size
        self.num_images = num_images

    @property
    def num_batches(self):
        return int(np.floor(self.num_images / self.batch_size))

    def __len__(self) -> int:
        """Returns the total number of batches."""
        return self.num_batches

    def __getitem__(self, idx: int) -> tuple[np.ndarray, np.ndarray]:
        """Generates one batch of random data"""
        # Generate random input and output data
        inputs = np.random.rand(
            self.batch_size, self.input_height, self.input_width, self.input_frames
        )
        outputs = np.random.rand(
            self.batch_size, self.output_height, self.output_width, self.output_frames
        )

        return inputs, outputs



import functools
from typing import List
import pandas as pd
import xarray as xr


def validate_data_lhs(func):
    """
    Decorator to validate data in LHS class fit method.

    Parameters
    ----------
    func : callable
        The function to be decorated

    Returns
    -------
    callable
        The decorated function
    """

    @functools.wraps(func)
    def wrapper(
        self,
        dimensions_names: List[str],
        lower_bounds: List[float],
        upper_bounds: List[float],
        num_samples: int,
    ):
        if not isinstance(dimensions_names, list):
            raise TypeError("Dimensions names must be a list")
        if not isinstance(lower_bounds, list):
            raise TypeError("Lower bounds must be a list")
        if not isinstance(upper_bounds, list):
            raise TypeError("Upper bounds must be a list")
        if len(dimensions_names) != len(lower_bounds) or len(lower_bounds) != len(
            upper_bounds
        ):
            raise ValueError(
                "Dimensions names, lower bounds and upper bounds must have the same length"
            )
        if not all(
            [lower <= upper for lower, upper in zip(lower_bounds, upper_bounds)]
        ):
            raise ValueError("Lower bounds must be less than or equal to upper bounds")
        if not isinstance(num_samples, int) or num_samples <= 0:
            raise ValueError("Variable num_samples must be integer and > 0")
        return func(self, dimensions_names, lower_bounds, upper_bounds, num_samples)

    return wrapper


def validate_data_mda(func):
    """
    Decorator to validate data in MDA class fit method.

    Parameters
    ----------
    func : callable
        The function to be decorated

    Returns
    -------
    callable
        The decorated function
    """

    @functools.wraps(func)
    def wrapper(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        custom_scale_factor: dict = {},
        first_centroid_seed: int = None,
    ):
        if data is None:
            raise ValueError("Data cannot be None")
        elif not isinstance(data, pd.DataFrame):
            raise TypeError("Data must be a pandas DataFrame")
        if not isinstance(directional_variables, list):
            raise TypeError("Directional variables must be a list")
        if not isinstance(custom_scale_factor, dict):
            raise TypeError("Custom scale factor must be a dict")
        if first_centroid_seed is not None:
            if (
                not isinstance(first_centroid_seed, int)
                or first_centroid_seed < 0
                or first_centroid_seed > data.shape[0]
            ):
                raise ValueError(
                    "First centroid seed must be an integer >= 0 and < num of data points"
                )
        return func(self, data, directional_variables, custom_scale_factor)

    return wrapper


def validate_data_kma(func):
    """
    Decorator to validate data in KMA class fit method.

    Parameters
    ----------
    func : callable
        The function to be decorated

    Returns
    -------
    callable
        The decorated function
    """

    @functools.wraps(func)
    def wrapper(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        custom_scale_factor: dict = {},
    ):
        if data is None:
            raise ValueError("Data cannot be None")
        elif not isinstance(data, pd.DataFrame):
            raise TypeError("Data must be a pandas DataFrame")
        if not isinstance(directional_variables, list):
            raise TypeError("Directional variables must be a list")
        if not isinstance(custom_scale_factor, dict):
            raise TypeError("Custom scale factor must be a dict")
        return func(self, data, directional_variables, custom_scale_factor)

    return wrapper


def validate_data_som(func):
    """
    Decorator to validate data in SOM class fit method.

    Parameters
    ----------
    func : callable
        The function to be decorated

    Returns
    -------
    callable
        The decorated function
    """

    @functools.wraps(func)
    def wrapper(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        num_iteration: int = 1000,
    ):
        if data is None:
            raise ValueError("Data cannot be None")
        elif not isinstance(data, pd.DataFrame):
            raise TypeError("Data must be a pandas DataFrame")
        if not isinstance(directional_variables, list):
            raise TypeError("Directional variables must be a list")
        if not isinstance(num_iteration, int) or num_iteration <= 0:
            raise ValueError("Number of iterations must be integer and > 0")
        return func(self, data, directional_variables, num_iteration)

    return wrapper


def validate_data_pca(func):
    """
    Decorator to validate data in PCA class fit method.

    Parameters
    ----------
    func : callable
        The function to be decorated

    Returns
    -------
    callable
        The decorated function
    """

    @functools.wraps(func)
    def wrapper(
        self,
        data: xr.Dataset,
        vars_to_stack: List[str],
        coords_to_stack: List[str],
        pca_dim_for_rows: str,
        window_in_pca_dim_for_rows: List[int] = [0],
        value_to_replace_nans: float = None,
    ):
        if data is None:
            raise ValueError("Data cannot be None")
        elif not isinstance(data, xr.Dataset):
            raise TypeError("Data must be an xarray Dataset")
        # Check that all vars_to_stack are in the data
        if not isinstance(vars_to_stack, list) or len(vars_to_stack) == 0:
            raise ValueError("Variables to stack must be a non-empty list")
        for var in vars_to_stack:
            if var not in data.data_vars:
                raise ValueError(f"Variable {var} not found in data")
        # Check that all variables in vars_to_stack have the same coordinates and dimensions
        first_var = vars_to_stack[0]
        first_var_dims = list(data[first_var].dims)
        first_var_coords = list(data[first_var].coords)
        for var in vars_to_stack:
            if list(data[var].dims) != first_var_dims:
                raise ValueError(
                    f"All variables must have the same dimensions. Variable {var} does not match."
                )
            if list(data[var].coords) != first_var_coords:
                raise ValueError(
                    f"All variables must have the same coordinates. Variable {var} does not match."
                )
        # Check that all coords_to_stack are in the data
        if not isinstance(coords_to_stack, list) or len(coords_to_stack) == 0:
            raise ValueError("Coordinates to stack must be a non-empty list")
        for coord in coords_to_stack:
            if coord not in data.coords:
                raise ValueError(f"Coordinate {coord} not found in data.")
        # Check that pca_dim_for_rows is in the data, and window > 0 if provided
        if not isinstance(pca_dim_for_rows, str) or pca_dim_for_rows not in data.dims:
            raise ValueError(
                "PCA dimension for rows must be a string and found in the data dimensions"
            )
        if window_in_pca_dim_for_rows is not None:
            if (
                not isinstance(window_in_pca_dim_for_rows, list)
                or len(window_in_pca_dim_for_rows) == 0
            ):
                raise ValueError(
                    "Window in PCA dimension for rows must be a non-empty list"
                )
        if value_to_replace_nans is not None:
            if not isinstance(value_to_replace_nans, float):
                raise ValueError("Value to replace NaNs must be float")
        return func(
            self,
            data,
            vars_to_stack,
            coords_to_stack,
            pca_dim_for_rows,
            window_in_pca_dim_for_rows,
            value_to_replace_nans,
        )

    return wrapper


def validate_data_rbf(func):
    """
    Decorator to validate data in RBF class fit method.

    Parameters
    ----------
    func : callable
        The function to be decorated

    Returns
    -------
    callable
        The decorated function
    """

    @functools.wraps(func)
    def wrapper(
        self,
        subset_data: pd.DataFrame,
        target_data: pd.DataFrame,
        subset_directional_variables: List[str] = [],
        target_directional_variables: List[str] = [],
        subset_custom_scale_factor: dict = {},
        normalize_target_data: bool = True,
        target_custom_scale_factor: dict = {},
        num_threads: int = None,
        iteratively_update_sigma: bool = False,
    ):
        if subset_data is None:
            raise ValueError("Subset data cannot be None")
        elif not isinstance(subset_data, pd.DataFrame):
            raise TypeError("Subset data must be a pandas DataFrame")
        if target_data is None:
            raise ValueError("Target data cannot be None")
        elif not isinstance(target_data, pd.DataFrame):
            raise TypeError("Target data must be a pandas DataFrame")
        if not isinstance(subset_directional_variables, list):
            raise TypeError("Subset directional variables must be a list")
        for directional_variable in subset_directional_variables:
            if directional_variable not in subset_data.columns:
                raise ValueError(
                    f"Directional variable {directional_variable} not found in subset data"
                )
        if not isinstance(target_directional_variables, list):
            raise TypeError("Target directional variables must be a list")
        for directional_variable in target_directional_variables:
            if directional_variable not in target_data.columns:
                raise ValueError(
                    f"Directional variable {directional_variable} not found in target data"
                )
        if not isinstance(subset_custom_scale_factor, dict):
            raise TypeError("Subset custom scale factor must be a dict")
        if not isinstance(normalize_target_data, bool):
            raise TypeError("Normalize target data must be a bool")
        if not isinstance(target_custom_scale_factor, dict):
            raise TypeError("Target custom scale factor must be a dict")
        if num_threads is not None:
            if not isinstance(num_threads, int) or num_threads <= 0:
                raise ValueError("Number of threads must be integer and > 0")
        if not isinstance(iteratively_update_sigma, bool):
            raise TypeError("Iteratively update sigma must be a boolean")
        return func(
            self,
            subset_data,
            target_data,
            subset_directional_variables,
            target_directional_variables,
            subset_custom_scale_factor,
            normalize_target_data,
            target_custom_scale_factor,
            num_threads,
            iteratively_update_sigma,
        )

    return wrapper



import os
from datetime import datetime
import pytz
import logging


def get_file_logger(
    name: str, logs_path: str = None, level: int = logging.INFO
) -> logging.Logger:
    """
    Creates and returns a logger that writes log messages to a file.

    Parameters
    ----------
    name : str
        The name of the logger.
    logs_path : str, optional
        The file path where the log messages will be written (default is None).
    level : int, optional
        The logging level (default is logging.INFO).

    Returns
    -------
    logger: logging.Logger
        Configured logger instance.

    Examples
    --------
    >>> from bluemath_tk.core.logging import get_file_logger
    >>> # Create a logger that writes to "app.log"
    >>> logger = get_file_logger("my_app_logger", "app.log")
    >>> # Log messages
    >>> logger.info("This is an info message.")
    >>> logger.warning("This is a warning message.")
    >>> logger.error("This is an error message.")
    >>> # The output will be saved in "app.log" with the format:
    >>> # 2023-10-22 14:55:23,456 - my_app_logger - INFO - This is an info message.
    >>> # 2023-10-22 14:55:23,457 - my_app_logger - WARNING - This is a warning message.
    >>> # 2023-10-22 14:55:23,458 - my_app_logger - ERROR - This is an error message.
    """

    # If a logger with the specified name already exists, return it
    if name in logging.Logger.manager.loggerDict:
        return logging.getLogger(name)

    # Create a logger with the specified name
    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.propagate = False  # Avoid duplicate logs

    # Get current date to append to logs_path
    date_str = datetime.now(pytz.timezone("Europe/Madrid")).strftime("%Y-%m-%d")

    # Create a file handler to write logs to the specified file
    if logs_path is None:
        os.makedirs("logs", exist_ok=True)
        logs_path = os.path.join("logs", f"{name.strip()}_{date_str}.log")
    else:
        os.makedirs(os.path.dirname(logs_path))
    file_handler = logging.FileHandler(logs_path)

    # Also ouput logs in the console
    console_handler = logging.StreamHandler()

    # Define a logging format
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setFormatter(formatter)
    console_handler.setFormatter(formatter)

    # Add the file handler to the logger
    logger.addHandler(file_handler)
    logger.addHandler(console_handler)

    return logger



import os
import itertools
from abc import abstractmethod
from typing import List
import subprocess
import numpy as np
from jinja2 import Environment, FileSystemLoader
from ..core.models import BlueMathModel


class BaseModelWrapper(BlueMathModel):
    """
    Base class for numerical models wrappers.

    Attributes
    ----------
    templates_dir : str
        The directory where the templates are stored.
    templates_name : List[str]
        The names of the templates.
    model_parameters : dict
        The parameters to be used in the templates.
    output_dir : str
        The directory where the output files will be saved.
    env : Environment
        The Jinja2 environment.
    cases_dirs : List[str]
        The list with cases directories.
    cases_context : List[dict]
        The list with cases context.

    Methods
    -------
    _check_parameters_type(default_parameters, model_parameters)
        Check if the parameters have the correct type.
    _exec_bash_commands(str_cmd, out_file=None, err_file=None)
        Execute bash commands.
    create_cases_context_one_by_one()
        Create an array of dictionaries with the combinations of values from the
        input dictionary, one by one.
    create_cases_context_all_combinations()
        Create an array of dictionaries with each possible combination of values
        from the input dictionary.
    render_file_from_template(template_name, context, output_filename=None)
        Render a file from a template.
    write_array_in_file(array, filename)
        Write an array in a file.
    copy_files(src, dst)
        Copy file(s) from source to destination.
    build_cases(mode="all_combinations")
        Create the cases folders and render the input files.
    run_cases()
        Run the cases.
    run_model(case_dir)
        Run the model for a specific case (abstract method).
    """

    def __init__(
        self,
        templates_dir: str,
        templates_name: List[str],
        model_parameters: dict,
        output_dir: str,
        default_parameters: dict = None,
    ) -> None:
        """
        Initialize the BaseModelWrapper.

        Parameters
        ----------
        templates_dir : str
            The directory where the templates are stored.
        templates_name : list
            The names of the templates.
        model_parameters : dict
            The parameters to be used in the templates.
        output_dir : str
            The directory where the output files will be saved.
        default_parameters : dict, optional
            The default parameters type for the model. If None, the parameters will
            not be checked.
            Default is None.
        """

        super().__init__()
        if default_parameters is not None:
            self._check_parameters_type(
                default_parameters=default_parameters, model_parameters=model_parameters
            )
        self.templates_dir = templates_dir
        self.templates_name = templates_name
        self.model_parameters = model_parameters
        self.output_dir = output_dir
        self._env = Environment(loader=FileSystemLoader(self.templates_dir))
        self.cases_dirs: List[str] = []
        self.cases_context: List[dict] = []

    @property
    def env(self) -> Environment:
        return self._env

    def _check_parameters_type(
        self, default_parameters: dict, model_parameters: dict
    ) -> None:
        """
        Check if the parameters have the correct type.
        This function is called in the __init__ method of the BaseModelWrapper,
        but default_parameters are defined in the child classes.
        This way, child classes can define default types for parameters.

        Parameters
        ----------
        default_parameters : dict
            The default parameters type for the model.
        model_parameters : dict
            The parameters to be used in the templates.

        Raises
        ------
        ValueError
            If a parameter has the wrong type.
        """

        for model_param, param_value in model_parameters.items():
            if model_param not in default_parameters:
                self.logger.warning(
                    f"Parameter {model_param} is not in the default_parameters"
                )
            else:
                if isinstance(param_value, (list, np.ndarray)) and all(
                    isinstance(item, default_parameters[model_param])
                    for item in param_value
                ):
                    self.logger.info(
                        f"Parameter {model_param} has the correct type: {default_parameters[model_param]}"
                    )
                else:
                    raise ValueError(
                        f"Parameter {model_param} has the wrong type: {default_parameters[model_param]}"
                    )

    @staticmethod
    def _exec_bash_commands(
        str_cmd: str, out_file: str = None, err_file: str = None
    ) -> None:
        """
        Execute bash commands.

        Parameters
        ----------
        str_cmd : str
            The bash command.
        out_file : str, optional
            The name of the output file. If None, the output will be printed in the terminal.
            Default is None.
        err_file : str, optional
            The name of the error file. If None, the error will be printed in the terminal.
            Default is None.
        """

        _stdout = None
        _stderr = None

        if out_file:
            _stdout = open(out_file, "w")
        if err_file:
            _stderr = open(err_file, "w")

        s = subprocess.Popen(str_cmd, shell=True, stdout=_stdout, stderr=_stderr)
        s.wait()

        if out_file:
            _stdout.flush()
            _stdout.close()
        if err_file:
            _stderr.flush()
            _stderr.close()

    def create_cases_context_one_by_one(self) -> List[dict]:
        """
        Create an array of dictionaries with the combinations of values from the
        input dictionary, one by one.

        Returns
        -------
        array_of_contexts : list
            A list of dictionaries, each representing a unique combination of
            parameter values.
        """

        num_cases = len(next(iter(self.model_parameters.values())))
        array_of_contexts = []
        for param, values in self.model_parameters.items():
            if len(values) != num_cases:
                raise ValueError(
                    f"All parameters must have the same number of values in one_by_one mode, check {param}"
                )

        for case_num in range(num_cases):
            case_context = {
                param: values[case_num]
                for param, values in self.model_parameters.items()
            }
            array_of_contexts.append(case_context)

        return array_of_contexts

    def create_cases_context_all_combinations(self) -> List[dict]:
        """
        Create an array of dictionaries with each possible combination of values
        from the input dictionary.

        Returns
        -------
        array_of_contexts : list
            A list of dictionaries, each representing a unique combination of
            parameter values.
        """

        keys = self.model_parameters.keys()
        values = self.model_parameters.values()
        combinations = itertools.product(*values)

        array_of_contexts = [
            dict(zip(keys, combination)) for combination in combinations
        ]

        return array_of_contexts

    def render_file_from_template(
        self, template_name: str, context: dict, output_filename: str = None
    ) -> None:
        """
        Render a file from a template.

        Parameters
        ----------
        template_name : str
            The name of the template file.
        context : dict
            The context to be used in the template.
        output_filename : str, optional
            The name of the output file. If None, it will be saved in the output
            directory with the same name as the template.
            Default is None.
        """

        template = self.env.get_template(name=template_name)
        rendered_content = template.render(context)
        if output_filename is None:
            output_filename = os.path.join(self.output_dir, template_name)
        with open(output_filename, "w") as f:
            f.write(rendered_content)

    def write_array_in_file(self, array: np.ndarray, filename: str) -> None:
        """
        Write an array in a file.

        Parameters
        ----------
        array : np.ndarray
            The array to be written. Can be 1D or 2D.
        filename : str
            The name of the file.
        """

        with open(filename, "w") as f:
            if array.ndim == 1:
                for item in array:
                    f.write(f"{item}\n")
            elif array.ndim == 2:
                for row in array:
                    f.write(" ".join(map(str, row)) + "\n")
            else:
                raise ValueError("Only 1D and 2D arrays are supported")

    def copy_files(self, src: str, dst: str) -> None:
        """
        Copy file(s) from source to destination.

        Parameters
        ----------
        src : str
            The source file.
        dst : str
            The destination file.
        """

        if os.path.isdir(src):
            os.makedirs(dst, exist_ok=True)
            for file in os.listdir(src):
                with open(file, "r") as f:
                    content = f.read()
                with open(os.path.join(dst, file), "w") as f:
                    f.write(content)
        else:
            with open(src, "r") as f:
                content = f.read()
            with open(dst, "w") as f:
                f.write(content)

    def build_cases(self, mode: str = "all_combinations") -> None:
        """
        Create the cases folders and render the input files.

        Parameters
        ----------
        mode : str, optional
            The mode to create the cases. Can be "all_combinations" or "one_by_one".
            Default is "all_combinations".
        """

        if mode == "all_combinations":
            self.cases_context = self.create_cases_context_all_combinations()
        elif mode == "one_by_one":
            self.cases_context = self.create_cases_context_one_by_one()
        else:
            raise ValueError(f"Invalid mode to create cases: {mode}")
        for case_num, case_context in enumerate(self.cases_context):
            case_context["case_num"] = case_num
            case_dir = os.path.join(self.output_dir, f"{case_num:04}")
            self.cases_dirs.append(case_dir)
            os.makedirs(case_dir, exist_ok=True)
            for template_name in self.templates_name:
                self.render_file_from_template(
                    template_name=template_name,
                    context=case_context,
                    output_filename=os.path.join(case_dir, template_name),
                )
        self.logger.info(
            f"{len(self.cases_dirs)} cases created in {mode} mode and saved in {self.output_dir}"
        )

    def run_cases(self) -> None:
        """
        Run the cases.
        """

        if self.cases_dirs:
            for case_dir in self.cases_dirs:
                self.logger.info(f"Running case in {case_dir}")
                self.run_model(case_dir=case_dir)
            self.logger.info("All cases ran successfully.")
        else:
            raise ValueError("No cases to run.")

    @abstractmethod
    def run_model(self, case_dir: str) -> None:
        """
        Run the model.

        Parameters
        ----------
        case_dir : str
            The directory of the case.
        """

        pass



from abc import abstractmethod
from ..core.models import BlueMathModel


class BaseDownloader(BlueMathModel):
    """
    Abstract class for BlueMath downloaders.

    Attributes
    ----------
    base_path_to_download : str
        The base path to download the data.
    debug : bool, optional
        If True, the logger will be set to DEBUG level. Default is True.
    check : bool, optional
        If True, just file checking is required. Default is False.

    Methods
    -------
    download_data(*args, **kwargs)
        Downloads the data. This method must be implemented in the child class.

    Notes
    -----
    - This class is an abstract class and should not be instantiated.
    - The download_data method must be implemented in the child class.
    """

    def __init__(
        self, base_path_to_download: str, debug: bool = True, check: bool = False
    ) -> None:
        """
        The constructor for BaseDownloader class.

        Parameters
        ----------
        base_path_to_download : str
            The base path to download the data.
        debug : bool, optional
            If True, the logger will be set to DEBUG level. Default is True.
        check : bool, optional
            If True, just file checking is required. Default is False.

        Raises
        ------
        ValueError
            If base_path_to_download is not a string.
            If debug is not a boolean.
            If check is not a boolean.

        Notes
        -----
        - The logger will be set to INFO level.
        - If debug is True, the logger will be set to DEBUG level.
        """

        super().__init__()
        if not isinstance(base_path_to_download, str):
            raise ValueError("base_path_to_download must be a string")
        self._base_path_to_download: str = base_path_to_download
        if not isinstance(debug, bool):
            raise ValueError("debug must be a boolean")
        self._debug: bool = debug
        if not isinstance(check, bool):
            raise ValueError("check must be a boolean")
        self._check: bool = check

    @property
    def base_path_to_download(self) -> str:
        return self._base_path_to_download

    @property
    def debug(self) -> bool:
        return self._debug

    @property
    def check(self) -> bool:
        return self._check

    @abstractmethod
    def download_data(self, *args, **kwargs) -> None:
        pass



from abc import abstractmethod
from typing import List
import pandas as pd
from ..core.models import BlueMathModel


class BaseInterpolation(BlueMathModel):
    """
    Base class for all interpolation BlueMath models.
    This class provides the basic structure for all interpolation models.

    Methods
    -------
    fit(*args, **kwargs)
    predict(*args, **kwargs)
    fit_predict(*args, **kwargs)
    """

    @abstractmethod
    def __init__(self):
        super().__init__()

    @abstractmethod
    def fit(self, *args, **kwargs):
        """
        Fits the model to the data.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.
        """

        pass

    @abstractmethod
    def predict(self, *args, **kwargs):
        """
        Predicts the interpolated data given a dataset.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.
        """

        pass

    @abstractmethod
    def fit_predict(self, *args, **kwargs):
        """
        Fits the model to the subset and predicts the interpolated dataset.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.
        """

        pass


class InterpolationComparator:
    """
    Class for comparing interpolation models.
    """

    def __init__(self, list_of_models: List[BaseInterpolation]) -> None:
        """
        Initializes the InterpolationComparator class.
        """

        self.list_of_models = list_of_models

    def fit(
        self,
        subset_data: pd.DataFrame,
        target_data: pd.DataFrame,
    ) -> None:
        """
        Fits the clustering models.
        """

        for model in self.list_of_models:
            model.fit(
                subset_data=subset_data,
                target_data=target_data,
            )



import keras
from models import resnet_model
from generators.ncDataGenerator import DataGenerator

# instantiate model class (load memory)
model = resnet_model.get_model(
    image_height=64, image_width=64, input_frames=1, output_frames=1
)

# print summary of the model
print(model.summary())

# instantiate generator class
train_generator = DataGenerator(
    msl_path="/home/tausiaj/DATA/Comparison-ERA5/msl_spain.nc",
    tp_path="/home/tausiaj/DATA/Comparison-ERA5/tp_spain.nc",
    num_images=8760,
    sequential=False,
    batch_size=1,
)

a, b = train_generator.__getitem__(1)
print(a.shape)
print(b.shape)
# define oprimizer
optimizer = keras.optimizers.AdamW
model.compile(
    optimizer=optimizer(learning_rate=1e-4, weight_decay=1e-5),
    loss=keras.losses.mean_squared_error,
)

# start the train loop with the fit method
history = model.fit(train_generator, initial_epoch=0, epochs=20, steps_per_epoch=500)


print("training complete")



import warnings
warnings.simplefilter('ignore')

import os
import os.path as op

import pickle

import glob
from datetime import datetime, timedelta, date
from cftime._cftime import DatetimeGregorian

import numpy as np
import pandas as pd
import xarray as xr

import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.colors import TwoSlopeNorm
import matplotlib.dates as mdates
import matplotlib.colors as mcolors

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import regionmask


from ..datamining.pca import PCA
from ..datamining.kma import KMA
from ..core.plotting.colors import get_config_variables, get_cluster_colors

config_variables = get_config_variables()

class DWT:

    '''
    This class implements the Princial Component Algorithm (PCA)
    
    '''

    def __init__(self):
        
        self.data = []
        self.p_data = []
        self.area = [0, -90, 360, 90]
        self.dim_pca = 'time'
        self

    def generate_dataset(self):
    
        p_data = self.p_data
        area = self.area
        resample_factor = self.resample_factor
        remove_land = self.remove_land
        estela = self.estela
        mslp_grad = self.mslp_grad
    
        for ivar, var in enumerate(self.predictor):
    
            p_file = glob.glob(os.path.join(p_data, '*{0}*1day*'.format(var)))[0]
            ds = xr.open_dataset(p_file)

            if var == 'sst':
                 if ds.sst.mean().values > 150:
                     ds['sst'] = ds['sst'] - 273.15 #transform kelvin to celsius

            if var == 'mslp':
                 if ds.mslp.mean().values > 10000:
                     ds['mslp'] = ds['mslp']/100 #transform Pa to HPa/mbar
            
    
            if not var in ds.data_vars.keys():
                raise DWTError("Variable {0} not found in dataset {1}. \
                Available vars: {2}. Rename variables to match file name".format(
                    var, p_file, list(ds.data_vars.keys())))
    
            expected_dims = ['time', 'longitude', 'latitude']
            correct_dims = all(dim in ds.dims for dim in expected_dims)
            if not correct_dims:
                print(ds.dims)
                raise DWTError("Incorrect dimensions found in dataset. Expected dimensions: {0}'".format(expected_dims))
    
            
            ds['longitude'] = (ds.longitude + 360) % 360 #Make sure longitudes are 0-360
            ds = ds.sortby(['longitude', 'latitude', 'time'])
            ds = ds.transpose('time', 'latitude', 'longitude')

            if area[1]<area[0]: #Area goes over 0
                ds = ds.sel(latitude = slice(area[2]-5, area[3]+5))
                ds = ds.isel(longitude = np.where((ds.longitude.values>area[0]-5) | (ds.longitude.values<area[1]+5))[0])
            else:
                ds = ds.sel(longitude = slice(area[0]-5, area[1]+5,),
                            latitude = slice(area[2]-5, area[3]+5))
    
            if ivar == 0:
                DS = ds.copy()
            else:
                ds = ds.interp_like(DS)
                DS[var] = (('time', 'latitude', 'longitude'), ds[var].values)
    
            #Coarsen
            if self.resample_factor>1:
                DS = DS.coarsen(longitude = resample_factor, 
                                latitude = resample_factor, 
                                boundary = 'pad').mean()

        self.data = DS

        
        #If mslp_grad calculate spatial sea level pressure gradients (proportional to wind)
        if mslp_grad:
            self.spatial_gradient(var_name = 'mslp')
        
        self.data = DS

        
        # Call dynamic_estela_predictor if estela is True
        if estela:
            DS = self.data
            ESTELA = xr.open_dataset('estela_sea.nc',decode_times=False)   # ESTELA file
            #ESTELA = ESTELA.sel(time='ALL')
            ESTELA = ESTELA.assign({'estela_mask':(('latitude','longitude'),np.where(ESTELA.F.values>0,1,np.nan))})
    
            # select ESTELA at site grid
            ESTELA_site = ESTELA.sortby(ESTELA.latitude,ascending=False).interp(coords={'latitude':DS.latitude,'longitude':DS.longitude})
            
            # apply ESTELA mask to SLP data
            DS['mslp'] = DS.mslp * ESTELA_site.estela_mask
            DS['mslp_grad'] = DS.mslp_grad * ESTELA_site.estela_mask
            DS['mask_land'] = ESTELA_site.estela_mask

            # generate estela predictor
            estela_D = ESTELA_site.traveltime # .drop('time')
            self.estela_D = estela_D
            self.data = DS
        
            self.data = self.dynamic_estela_predictor()




        #Remove land areas (if only ocean is of interest)
        if remove_land:
            
            mask = 1 - regionmask.defined_regions.natural_earth_v4_1_0.land_110.mask(
                    DS, lon_name = 'longitude', lat_name = 'latitude').fillna(1).rename("lsm")
    
            DS = DS.where(mask.where(mask == 0) == 0)        

        
        if area[1]<area[0]: #Area goes over 0
            DS = DS.sel(latitude = slice(area[2]-5, area[3]+5))
            DS = DS.isel(longitude = np.where((DS.longitude.values>area[0]) | (DS.longitude.values<area[1]))[0])
            #order from -180 to 180
            DS['longitude'] = np.where(DS.longitude.values>180, DS.longitude.values - 360, DS.longitude.values)
            DS = DS.sortby('longitude')
        
        else:
            DS = DS.sel(longitude = slice(area[0]-5, area[1]+5,),
                        latitude = slice(area[2]-5, area[3]+5))
        



    def dynamic_estela_predictor(self):
        '''
        Generate dynamic predictor using estela
    
        xdset:
            (time, latitude, longitude), var_name, mask
    
        returns similar xarray.Dataset with variables:
            (time, latitude, longitude), var_name_comp
            (time, latitude, longitude), var_name_gradient_comp
        '''

        xdset = self.data
        var_name = 'mslp'
        estela_D = self.estela_D 


        # first day is estela max
        first_day = int(np.floor(np.nanmax(estela_D)))+1
    
        # output will start at time=first_day
        shp = xdset[var_name].shape
        comp_shape = (shp[0]-first_day, shp[1], shp[2])
        var_comp = np.ones(comp_shape) * np.nan
        var_grd_comp = np.ones(comp_shape) * np.nan
    
        # get data using estela for each cell
        for i_lat in range(len(xdset.latitude)):
            for i_lon in range(len(xdset.longitude)):
                ed = estela_D[i_lat, i_lon]
                if not np.isnan(ed):
    
                    # mount estela displaced time array
                    i_times = np.arange(
                        first_day, len(xdset.time)
                    ) - int(ed) # np.int
    
                    # select data from displaced time array positions
                    xdselec = xdset.isel(
                        time = i_times,
                        latitude = i_lat,
                        longitude = i_lon)
    
                    # get estela predictor values
                    var_comp[:, i_lat, i_lon] = xdselec[var_name].values
                    var_grd_comp[:, i_lat, i_lon] = xdselec['{0}_grad'.format(var_name)].values
    
        # return generated estela predictor
        return xr.Dataset(
            {
                '{0}'.format(var_name):(
                    ('time','latitude','longitude'), var_comp),
                '{0}_grad'.format(var_name):(
                    ('time','latitude','longitude'), var_grd_comp),
    
            },
            coords = {
                'time':xdset.time.values[first_day:],
                'latitude':xdset.latitude.values,
                'longitude':xdset.longitude.values,
            }
        )



    def spatial_gradient(self, var_name):
        
        '''
        Calculate spatial gradient
    
        xdset:
            (time, 'latitude', 'longitude'), var_name
    
        returns xdset with new variable "var_name_gradient"
        '''
    
        ds = self.data

        var_val = ds[var_name].values
        var_grad = np.zeros_like(var_val)
    
        lat_phi = np.pi * np.abs(ds.latitude.values) / 180.0
    
        for it in range(len(ds.time)):
            m_c = var_val[it, 1:-1, 1:-1]
            m_l = var_val[it, 1:-1, :-2]
            m_r = var_val[it, 1:-1, 2:]
            m_u = var_val[it, :-2, 1:-1]
            m_d = var_val[it, 2:, 1:-1]
    
            dpx1 = (m_c - m_l) / np.cos(lat_phi[None, None, 1:-1, None])
            dpx2 = (m_r - m_c) / np.cos(lat_phi[None, None, 1:-1, None])
            dpy1 = m_c - m_d
            dpy2 = m_u - m_c
    
            vg = (dpx1**2 + dpx2**2) / 2 + (dpy1**2 + dpy2**2) / 2
            var_grad[it, 1:-1, 1:-1] = vg.squeeze()
    
        ds[f'{var_name}_grad'] = (('time', 'latitude', 'longitude'), var_grad)
        self.data = ds[sorted(ds.data_vars)]


        
    def pca(self, n_data = 10, n_pca=5):
    
        pca_ob = PCA()
        pca_ob.dataset = self.data
        pca_ob.dim_pca = self.dim_pca # coordinate to do PCA over
        pca_ob.generate_data_matrix()
        pca_ob.area = self.area
    
        cp = {
               'cmap':'RdBu_r'
                }
        
        pca_ob.plot_data(N=n_data, custom_params = cp)
    
        pca_ob.pca()
    
        pca_ob.plot_eofs_pcs(n_pca = n_pca)

        #TODO: decide wether pca_ob is too much or not
        self.pca_ob = pca_ob
        #self.pca = pca_ob.pca
        
    def sort_bmus(self):
    
        if 'mslp' in list(self.dwt_centroids.data_vars.keys()):
            var_order = 'mslp'
        else:
            var_order = list(self.dwt_centroids.data_vars.keys())[0]
           
        order = np.argsort(self.dwt_centroids[var_order].mean(dim = ('longitude', 'latitude')).values)
        print(order)
    
        data = self.data
        mapping = {val: i for i, val in enumerate(order)}
        v_ordered = [mapping[val] for val in data.bmus.values]
        data['bmus'] = (('time'), v_ordered)
    
        centroids = self.dwt_centroids
        centroids = centroids.isel(bmus = order)
        centroids['bmus'] = (('bmus'), range(len(centroids.bmus)))
        self.dwt_centroids = centroids.sortby('bmus')

    def kma(self, sort_bmus = True):
        
        kma_ob = KMA()
        kma_ob.n_pcs = np.where(self.pca_ob.pca.APEV <= self.explained_variance)[0][-1]
    
        kma_ob.data = self.pca_ob.pca.PCs.isel(n_components = range(kma_ob.n_pcs)).values
        kma_ob.ix_scalar = range(kma_ob.data.shape[1])  # Scalar columns
        kma_ob.ix_directional = [] #Directional columns 
        
        cu, cont = 0, 0
        while np.nanmin(cu) < self.min_data:
            
            kma_ob.kma(self.n_dwt)
            u, cu = np.unique(kma_ob.bmus, return_counts=True)
            cont+=1
            
        kma_ob.num_iter = cont
    
        self.kma_ob = kma_ob
    
        self.data['bmus'] = (('time'), self.kma_ob.bmus)
        self.dwt_centroids = self.data.groupby('bmus').mean(dim='time')
    
        if sort_bmus:
            self.sort_bmus()
            
    
        
    ### Plotting ###  

    def plot_map_features(self, ax, land_color = cfeature.COLORS['land']):
        
        ax.add_feature(cfeature.COASTLINE)
        ax.add_feature(cfeature.BORDERS, linestyle=':')
        ax.add_feature(cfeature.LAND, edgecolor='black', color = land_color)
        ax.add_feature(cfeature.OCEAN, color='lightblue', alpha = .5)

    def plot_map_mean(self):

        ds = self.data
        p_site = self.p_site
        vars = list(ds.data_vars.keys())
        area = self.area
        estela = self.estela
        
        if area[1]<area[0]:
            central_longitude = 0
        else:
            central_longitude = 180
    
    
        if len(vars)>3:
            nrows = ncols = int(np.ceil(np.sqrt(len(vars))))
        else:
            ncols = len(vars); nrows = 1
        
        fig, axs = plt.subplots(nrows, ncols, figsize = [15, 7], 
                                subplot_kw={'projection': ccrs.PlateCarree(central_longitude = central_longitude)})
        
        plt.subplots_adjust(left=0.02, right=0.98, top=0.95, bottom=0.05, wspace=0.05, hspace=0.05)
        
        if (nrows==1) & (ncols == 1):
            unique_ax = True 
        else:
            axs_list = axs.flatten()
            unique_ax = False
        
        for iv, var in enumerate(vars):

            if not var in config_variables.keys():
                config_variables[var] = config_variables['other']
        
            if unique_ax:
                ax = axs
            else:
                ax = axs_list[iv]
            
            self.plot_map_features(ax)

            vmin, vmax = config_variables[var]['limits'][0], config_variables[var]['limits'][1]
            
            contour = ax.contourf(ds.longitude, ds.latitude, ds[var].mean(dim = 'time'), 50,
                                  cmap=config_variables[var]['cmap'], transform=ccrs.PlateCarree(), 
                                  vmin = vmin, vmax = vmax)

            if not estela: 
                ax.set_extent([ds.longitude.values.min()-10, ds.longitude.values.max()+10, 
                              ds.latitude.values.min()-10, ds.latitude.values.max()+10, ], crs=ccrs.PlateCarree())
                
            else:                    
                if area[1]<area[0]: #Area goes over 0:
                    ax.set_extent([area[0]-360-5, area[1]+5, 
                                  area[2]-10, area[3]+10, ], crs=ccrs.PlateCarree())
                else: 
                    ax.set_extent([area[1]-10, area[0]+10, 
                                  area[2]-10, area[3]+10, ], crs=ccrs.PlateCarree())
            
                

            plt.colorbar(contour, ax = ax, orientation = 'horizontal', shrink = .7)
            ax.set_title(var)

        fig.savefig(op.join(p_site, 'plot_map_mean.png'), bbox_inches='tight')

        with open(op.join(p_site, 'plot_map_mean.pkl'), 'wb') as f:
            pickle.dump(fig, f)



    def plot_dwts(self, anomaly = False):   

        data_means = self.data.mean(dim = 'time')
        ds = self.dwt_centroids
        vars = list(ds.data_vars.keys())
        n_dwts = len(self.dwt_centroids.bmus.values)

        p_site = self.p_site
        
        area = self.area
        if area[1]<area[0]:
            central_longitude = 0
        else:
            central_longitude = 180
        
        if n_dwts>3:
            nrows = ncols = int(np.ceil(np.sqrt(n_dwts)))
        else:
            ncols = n_dwts; nrows = 1

        
        for iv, var in enumerate(vars):  
        
            fig, axs = plt.subplots(nrows, ncols, figsize = [20, 18/(len(ds.longitude.values)/len(ds.latitude.values))], 
                                subplot_kw={'projection': ccrs.PlateCarree(central_longitude = central_longitude)})
            
            if (nrows==1) & (ncols == 1):
                unique_ax = True 
            else:
                axs_list = axs.flatten()
                unique_ax = False
        
            plt.subplots_adjust(left=0.02, right=0.98, top=0.92, bottom=0.01, wspace=0.05, hspace=0.05)
        
            if not var in config_variables.keys():
                config_variables[var] = config_variables['other']

            norm = None
            if anomaly:
                total = ds[var].values - data_means[var].values
                lim = np.nanmax([np.abs(np.nanpercentile(total, 1)), np.abs(np.nanpercentile(total, 99))])
                vmin, vmax = -lim, lim
                cmap = 'RdBu_r'
            else:
                #vmin, vmax = config_variables[var]['limits'][0], config_variables[var]['limits'][1]
                vmin, vmax = np.nanpercentile(ds[var],1), np.nanpercentile(ds[var],99)
                if var == 'mslp':
                    norm = TwoSlopeNorm(vmin=vmin, vcenter=1014, vmax=vmax)
                cmap = config_variables[var]['cmap']

            np_colors_int = get_cluster_colors(n_dwts)
            
            for d in range(n_dwts):
                
                if unique_ax:
                    ax = axs
                else:
                    ax = axs_list[d]

                color_css = mcolors.to_hex(np.array([np.append(np_colors_int[d], .5) ]), keep_alpha=True)
                
                self.plot_map_features(ax, land_color = color_css)
        
                if anomaly:
                    var_plot = ds[var].isel(bmus = d) - data_means[var]
                else:
                    var_plot = ds[var].isel(bmus = d)

                if norm:
                    contour = ax.pcolor(ds.longitude, ds.latitude, var_plot, 
                                          cmap=cmap, transform=ccrs.PlateCarree(), 
                                          norm = norm)
                else:
                    contour = ax.pcolor(ds.longitude, ds.latitude, var_plot, 
                                          cmap=cmap, transform=ccrs.PlateCarree(), 
                                          vmin = vmin, vmax = vmax)

                # Add the number of each DWT
                ax.text(0.05, 0.05, d + 1, ha='left', va='bottom', fontsize=15, fontweight='bold', color='navy', transform=ax.transAxes)
            
            cbar = plt.colorbar(contour, ax = axs_list, orientation = 'vertical', shrink = .6)
            cbar.set_label(config_variables[var]['label'], fontsize = 16)





            fig.savefig(op.join(p_site, 'plot_dwts_{}_anomaly_{}.png'.format(var, anomaly)), bbox_inches='tight')
            with open(op.join(p_site, 'plot_dwts_{}_anomaly_{}.pkl'.format(var, anomaly)), 'wb') as f:
                pickle.dump(fig, f)
            
            plt.show()
            print('\n\n')
            

    def get_years_months_days(self, time):
        '''
        Returns years, months, days of time in separete lists
    
        (Used to avoid problems with dates type)
        '''
    
        t0 = time[0]
        if isinstance(t0, (date, datetime, DatetimeGregorian)):
            ys = np.asarray([x.year for x in time])
            ms = np.asarray([x.month for x in time])
            ds = np.asarray([x.day for x in time])
    
        else:
            tpd = pd.DatetimeIndex(time)
            ys = tpd.year
            ms = tpd.month
            ds = tpd.day
    
        return ys, ms, ds
        
    def ClusterProbabilities(self, series, set_values):
        'return series probabilities for each item at set_values'
    
        us, cs = np.unique(series, return_counts=True)
        d_count = dict(zip(us,cs))
    
        # cluster probabilities
        cprobs = np.zeros((len(set_values)))
        for i, c in enumerate(set_values):
           cprobs[i] = 1.0*d_count[c]/len(series) if c in d_count.keys() else 0.0
    
        return cprobs


    def ClusterProbs_Month(self, bmus, time, wt_set, month_ix):
        'Returns Cluster probs by month_ix'
    
        # get months
        _, months, _ = self.get_years_months_days(time)
    
        if isinstance(month_ix, list):
    
            # get each month indexes
            l_ix = []
            for m_ix in month_ix:
                ixs = np.where(months == m_ix)[0]
                l_ix.append(ixs)
    
            # get all indexes     
            ix = np.unique(np.concatenate(tuple(l_ix)))
    
        else:
            ix = np.where(months == month_ix)[0]
    
        bmus_sel = bmus[ix]
    
        return self.ClusterProbabilities(bmus_sel, wt_set)

    def axplot_wt_probs(self, ax, wt_probs,
                         ttl = '', vmin = 0, vmax = 0.1,
                         cmap = 'Blues', caxis='black'):
        'axes plot WT cluster probabilities'
    
        # clsuter transition plot
        pc = ax.pcolor(
            np.flipud(wt_probs),
            cmap=cmap, vmin=vmin, vmax=vmax,
            edgecolors='k',
        )
    
        # customize axes
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_title(ttl, {'fontsize':10, 'fontweight':'bold'})
    
        # axis color
        plt.setp(ax.spines.values(), color=caxis)
        plt.setp(
            [ax.get_xticklines(), ax.get_yticklines()],
            color=caxis,
        )
    
        # axis linewidth
        if caxis != 'black':
            plt.setp(ax.spines.values(), linewidth=3)
    
        return pc

    def axplot_wt_hist(self, ax, bmus, n_clusters, ttl=''):
        'axes plot WT cluster count histogram'
    
        # cluster transition plot
        ax.hist(
            bmus,
            bins = np.arange(1, n_clusters+2),
            edgecolor='k'
        )
    
        # customize axes
        #ax.grid('y')
    
        ax.set_xticks(np.arange(1,n_clusters+1)+0.5)
        ax.set_xticklabels(np.arange(1,n_clusters+1))
        ax.set_xlim([1, n_clusters+1])
        ax.tick_params(axis='both', which='major', labelsize=6)
    
        ax.set_title(ttl, {'fontsize':10, 'fontweight':'bold'})
    
    

    def ClusterProbabilities(self, series, set_values):
        'return series probabilities for each item at set_values'
    
        us, cs = np.unique(series, return_counts=True)
        d_count = dict(zip(us,cs))
    
        # cluster probabilities
        cprobs = np.zeros((len(set_values)))
        for i, c in enumerate(set_values):
           cprobs[i] = 1.0*d_count[c]/len(series) if c in d_count.keys() else 0.0
    
        return cprobs


    def plot_dwts_probs(self, vmax = 0.15, vmax_seasonality = 0.15):
    
        '''
        Plot Daily Weather Types bmus probabilities
        '''
    
        bmus = self.data.bmus.values
        bmus_time = self.data.time.values
        n_clusters = len(self.dwt_centroids.bmus.values)
        p_site = self.p_site
    
        wt_set = np.arange(n_clusters) + 1
    
        # best rows cols combination
        if n_clusters>3:
            n_rows = n_cols = int(np.ceil(np.sqrt(n_clusters)))
        else:
            n_cols = n_clusters; n_rows = 1
    
        # figure
        fig = plt.figure(figsize=(15, 9))
    
        # layout
        gs = gridspec.GridSpec(4, 7, wspace=0.10, hspace=0.25)
    
        # list all plots params
        l_months = [
            (1, 'January',   gs[1,3]),
            (2, 'February',  gs[2,3]),
            (3, 'March',     gs[0,4]),
            (4, 'April',     gs[1,4]),
            (5, 'May',       gs[2,4]),
            (6, 'June',      gs[0,5]),
            (7, 'July',      gs[1,5]),
            (8, 'August',    gs[2,5]),
            (9, 'September', gs[0,6]),
            (10, 'October',  gs[1,6]),
            (11, 'November', gs[2,6]),
            (12, 'December', gs[0,3]),
        ]
    
        l_3months = [
            ([12, 1, 2],  'DJF', gs[3,3]),
            ([3, 4, 5],   'MAM', gs[3,4]),
            ([6, 7, 8],   'JJA', gs[3,5]),
            ([9, 10, 11], 'SON', gs[3,6]),
        ]
    
        # plot total probabilities
        c_T = self.ClusterProbabilities(bmus, wt_set)
        C_T = np.reshape(c_T, (n_rows, n_cols))
    
        ax_probs_T = plt.subplot(gs[:2, :2])
        pc = self.axplot_wt_probs(ax_probs_T, C_T, ttl = 'DWT Probabilities')
    
        # plot counts histogram
        ax_hist = plt.subplot(gs[2:, :3])
        self.axplot_wt_hist(ax_hist, bmus, n_clusters, ttl = 'DWT Counts')
    
        # plot probabilities by month
        
        for m_ix, m_name, m_gs in l_months:
    
            # get probs matrix
            c_M = self.ClusterProbs_Month(bmus, bmus_time, wt_set, m_ix)
            C_M = np.reshape(c_M, (n_rows, n_cols))
    
            # plot axes
            ax_M = plt.subplot(m_gs)
            self.axplot_wt_probs(ax_M, C_M, ttl = m_name, vmax=vmax)
    
    
        # plot probabilities by 3 month sets
        
        for m_ix, m_name, m_gs in l_3months:
    
            # get probs matrix
            c_M = self.ClusterProbs_Month(bmus, bmus_time, wt_set, m_ix)
            C_M = np.reshape(c_M, (n_rows, n_cols))
    
            # plot axes
            ax_M = plt.subplot(m_gs)
            self.axplot_wt_probs(ax_M, C_M, ttl = m_name, vmax=vmax_seasonality, cmap='Greens')
    
        # add custom colorbar
        pp = ax_probs_T.get_position()
        cbar_ax = fig.add_axes([pp.x1+0.02, pp.y0, 0.02, pp.y1 - pp.y0])
        cb = fig.colorbar(pc, cax=cbar_ax, cmap='Blues')
        cb.ax.tick_params(labelsize=8)

        fig.savefig(op.join(p_site, 'plot_dwts_probs.png'), bbox_inches='tight')
        with open(op.join(p_site, 'plot_dwts_probs.pkl'), 'wb') as f:
            pickle.dump(fig, f)

    def Generate_PerpYear_Matrix(self, num_clusters, bmus_values, bmus_dates,
                                 num_sim=1, month_ini=1):
        '''
        Calculates and returns matrix for stacked bar plotting
    
        bmus_dates - datetime.datetime (only works if daily resolution)
        bmus_values has to be 2D (time, nsim)
        '''
    
        # generate perpetual year list
        list_pyear = self.GenOneYearDaily(month_ini=month_ini)
    
        # generate aux arrays
        m_plot = np.zeros((num_clusters, len(list_pyear))) * np.nan
        bmus_dates_months = np.array([pd.to_datetime(d).month for d in bmus_dates])
        bmus_dates_days = np.array([pd.to_datetime(d).day for d in bmus_dates])
    
        # sort data
        for i, dpy in enumerate(list_pyear):
            _, s = np.where(
                [(bmus_dates_months == dpy.month) & (bmus_dates_days == dpy.day)]
            )
    
            b = bmus_values[s]
            b = b.flatten()
    
            for j in range(num_clusters):
                _, bb = np.where([(j+1 == b)])  # j+1 starts at 1 bmus value!
    
                m_plot[j,i] = float(len(bb)/float(num_sim))/len(s)
    
        return m_plot
    
    def GenOneYearDaily(self, yy=1981, month_ini=1):
        'returns one generic year in a list of datetimes. Daily resolution'
    
        dp1 = datetime(yy, month_ini, 1)
        dp2 = dp1+timedelta(days=365)
    
        return [dp1 + timedelta(days=i) for i in range((dp2-dp1).days)]


    def plot_perpetual_year(self):
    
        num_clusters = len(self.dwt_centroids.bmus)
        bmus_values = self.data.bmus.values
        bmus_dates = self.data.time.values
        month_ini = 1
        p_site = self.p_site
    
        'axes plot bmus perpetual year'
    
        # get cluster colors for stacked bar plot
        np_colors_int = get_cluster_colors(num_clusters)


        # generate dateticks
        x_val = self.GenOneYearDaily(month_ini = month_ini)
    
        # generate plot matrix
        m_plot = self.Generate_PerpYear_Matrix(
            num_clusters, bmus_values + 1, bmus_dates,
            month_ini = month_ini)
    
        fig, ax = plt.subplots(1, figsize = (15, 5))
    
        # plot stacked bars
        bottom_val = np.zeros(m_plot[1,:].shape)
        for r in range(num_clusters):
            row_val = m_plot[r,:]
            ax.bar(
                x_val, row_val, bottom = bottom_val,
                width = 1, color = np.array([np_colors_int[r]]),
                alpha = .7
            )
    
            # store bottom
            bottom_val += row_val
    
        # customize  axis
        months = mdates.MonthLocator()
        monthsFmt = mdates.DateFormatter('%b')
    
        ax.set_xlim(x_val[0], x_val[-1])
        ax.xaxis.set_major_locator(months)
        ax.xaxis.set_major_formatter(monthsFmt)
        ax.set_ylim(0, 1)
        ax.set_ylabel('')

        fig.savefig(op.join(p_site, 'plot_perpetual_year.png'), bbox_inches='tight')
        with open(op.join(p_site, 'plot_perpetual_year.pkl'), 'wb') as f:
            pickle.dump(fig, f)


class DWTError(Exception):
    """Custom exception for DWT class."""
    def __init__(self, message="DWT error occurred."):
        self.message = message
        super().__init__(self.message)


import sys
from .._base_wrappers import BaseModelWrapper


class SwanModelWrapper(BaseModelWrapper):
    """
    Wrapper for the SWAN model.
    https://swanmodel.sourceforge.io/online_doc/swanuse/swanuse.html

    Attributes
    ----------
    swan_exec : str
        The SWAN executable path.
    default_parameters : dict
        The default parameters type for the model.

    Methods
    -------
    set_swan_exec(swan_exec: str) -> None
        Set the SWAN executable path.
    run_model(case_dir: str, log_file: str = "swan_exec.log") -> None
        Run the SWAN model for the specified case.
    """

    default_parameters = {
        "hs": float,
        "tp": float,
        "dir": float,
        "spr": float,
    }

    def __init__(
        self,
        templates_dir: str,
        templates_name: dict,
        model_parameters: dict,
        output_dir: str,
    ):
        """
        Initialize the SWAN model wrapper.

        Parameters
        ----------
        templates_dir : str
            The directory where the templates are stored.
        templates_name : list
            The names of the templates.
        model_parameters : dict
            The parameters to be used in the templates.
        output_dir : str
            The directory where the output files will be saved.
        """

        super().__init__(
            templates_dir=templates_dir,
            templates_name=templates_name,
            model_parameters=model_parameters,
            output_dir=output_dir,
            default_parameters=self.default_parameters,
        )
        self.set_logger_name(self.__class__.__name__)
        self._swan_exec: str = None

    @property
    def swan_exec(self) -> str:
        return self._swan_exec

    def set_swan_exec(self, swan_exec: str) -> None:
        self._swan_exec = swan_exec

    def run_model(self, case_dir: str, log_file: str = "swan_exec.log") -> None:
        """
        Run the SWAN model for the specified case.

        Parameters
        ----------
        case_dir : str
            The case directory.
        log_file : str, optional
            The log file name. Default is "swan_exec.log".

        Raises
        ------
        ValueError
            If the SWAN executable was not set.
        """

        if not self.swan_exec:
            raise ValueError("The SWAN executable was not set.")
        # check if windows OS
        is_win = sys.platform.startswith("win")
        if is_win:
            cmd = "cd {0} && {1} input".format(case_dir, self.swan_exec)
        else:
            cmd = "cd {0} && {1} -input input.sws".format(case_dir, self.swan_exec)
        # redirect output
        cmd += f" 2>&1 > {log_file}"
        # execute command
        self._exec_bash_commands(str_cmd=cmd)



import os
import numpy as np
from bluemath_tk.wrappers.xbeach import XBeachModelWrapper


class MyXBeachModelWrapper(XBeachModelWrapper):
    def build_cases(
        self,
        mode: str = "all_combinations",
        grd: str = None,
        bathy: np.ndarray = None,
        friction: np.ndarray = None,
    ):
        # Call the base class method to retain the original functionality
        super().build_cases(mode=mode)
        # Create the cases folders and render the input files
        if not self.cases_context or not self.cases_dirs:
            raise ValueError("Cases were not properly built.")
        for case_context, case_dir in zip(self.cases_context, self.cases_dirs):
            if grd is not None:
                # copy the grd file to the case folder
                self.copy_files(
                    src=grd, dst=os.path.join(case_dir, os.path.basename(grd))
                )
            if bathy is not None:
                # Save the bathymetry to a file
                self.write_array_in_file(
                    array=bathy, filename=os.path.join(case_dir, "bathy_000.dep")
                )
            if friction is not None:
                # Save the friction to a file
                self.write_array_in_file(
                    array=friction, filename=os.path.join(case_dir, "friction.txt")
                )


# Usage example
if __name__ == "__main__":
    # Define the input parameters
    templates_dir = (
        "/home/tausiaj/GitHub-GeoOcean/BlueMath/bluemath_tk/wrappers/xbeach/templates/"
    )
    templates_name = ["params.txt", "loclist.txt"]
    model_parameters = {
        "thetamax": [360],
        "dtheta": [4, 50, 6],
        "spectra": ["JONSWAP", "DIAZIN"],
    }
    output_dir = "/home/tausiaj/GitHub-GeoOcean/BlueMath/test_cases/xbeach/"
    # Create the bathymetry
    bathy = np.random.randn(100, 100)
    # Create the friction
    friction = np.random.randn(100, 100)
    # Create an instance of the XBEACH model wrapper
    swan_model = MyXBeachModelWrapper(
        templates_dir=templates_dir,
        templates_name=templates_name,
        model_parameters=model_parameters,
        output_dir=output_dir,
    )
    # Build the input files
    swan_model.build_cases(
        mode="all_combinations",
        grd="/home/tausiaj/Downloads/caso_XB/laredo_x5.grd",
        bathy=bathy,
        friction=friction,
    )
